#include <assert.h>

#include <pro.h>
#include <ua.hpp>

#include "interfaces/STARSTypes.h"

#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPInstr.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"
#include "interfaces/idapro/all.h"

using namespace std;

static uint32_t UseMacros[STARS_UA_MAXOP] = {STARS_CF_USE1, STARS_CF_USE2, STARS_CF_USE3, STARS_CF_USE4, STARS_CF_USE5, STARS_CF_USE6};

static uint32_t DefMacros[STARS_UA_MAXOP] = {STARS_CF_CHG1, STARS_CF_CHG2, STARS_CF_CHG3, STARS_CF_CHG4, STARS_CF_CHG5, STARS_CF_CHG6};

STARS_InstructionID_t STARS_IDA_Instruction_t::GetNextInstructionID(void) const {
	STARS_ea_t addr = this->m_id.GetIDWithinFile();

	SMPInstr TempInst(addr);
	if (!TempInst.FillCmd()) {
		return STARS_BADADDR;
	}
	STARS_ea_t next_addr = addr + TempInst.GetSize();

	return STARS_InstructionID_t(next_addr, this->m_id.GetFileNum());
}

STARS_InstructionID_t STARS_IDA_Instruction_t::GetTargetInstructionID(void) const {
	assert(NULL != this->STARScmd.Operands[0]);
	STARS_ea_t TargetAddr = this->STARScmd.Operands[0]->GetAddr();
	return STARS_InstructionID_t(TargetAddr, this->m_id.GetFileNum());
}

void STARS_IDA_Instruction_t::InitOperand(op_t &InitOp) const {
#if 0
	InitOp.n = 0;
	InitOp.type = o_void;
	InitOp.offb = 0;
	InitOp.offo = 0;
	InitOp.flags = 0;
	InitOp.set_showed();
	// NOTE: InitOp.dtyp field is initialized in IDAP_run() to 32 or 64 bits.
	InitOp.reg = R_none;
	InitOp.value = 0;
	InitOp.addr = 0;
	InitOp.specval = 0;
	InitOp.specflag1 = 0;
	InitOp.specflag2 = 0;
	InitOp.specflag3 = 0;
	InitOp.specflag4 = 0;
#else
	(void) memset(&InitOp, 0, sizeof(op_t));
	InitOp.dtyp = global_STARS_program->GetSTARS_ISA_dtyp();
#if IDA_SDK_VERSION < 680
	InitOp.set_showed();
#else
	InitOp.set_shown();
#endif
#endif

	return;
} // end of STARS_IDA_Instruction_t::InitOperand()

// Get instruction info by address from IDA Pro.
bool STARS_IDA_Instruction_t::STARS_GetCmd(void) {
	bool success = true;
	int InstrLen = 0;

	// Fill cmd structure with disassembly of instr
	InstrLen = decode_insn(m_id.GetIDWithinFile());

	// Copy cmd fields to member STARScmd.
	this->STARScmd.itype = cmd.itype;
	this->STARScmd.size = cmd.size;
	this->STARScmd.auxpref = cmd.auxpref;
	this->STARScmd.segpref = cmd.segpref;
	this->STARScmd.insnpref = cmd.insnpref;
	this->STARScmd.flags = cmd.flags;

	if (0 >= InstrLen) {
		SMP_msg("ERROR: decode_insn failed at %ullx in file %u\n", (unsigned long long) m_id.GetIDWithinFile(), m_id.GetFileNum());
		this->STARScmd.size = 0;
		for (int i = 0; i < STARS_UA_MAXOP; ++i) {
			this->STARScmd.Operands.push_back(nullptr);
			this->STARScmd.Operands[i] = this->MakeVoidOpnd();
		}
		success = false;
		return success;
	}

	// Get the canonical features into member STARSfeatures.
	this->STARSfeatures = cmd.get_canon_feature();

	for (int i = 0; i < STARS_UA_MAXOP; ++i) {
		this->STARScmd.Operands.push_back(std::make_shared<STARS_IDA_op_t>(cmd.Operands[i]));
		this->STARScmd.Operands[i]->SetSpecFlag4(0); 
#ifdef __EA64__
		if (global_STARS_program->GetSTARS_ISA_Bitwidth() == 64) {
			// Copy the cmd.rex prefix into the op_t.specflag4 field for each operand
			//  that has a SIB byte.
			this->STARScmd.Operands[i]->SetSpecFlag4(this->STARScmd.rex);
		}
#endif
		// See comments on STARS_VEXPR and STARS_VSIB in SMPDataFlowAnalysis.h.
		//  These bits do not (as of IDA Pro 6.4) conflict with cmd.rex bits.
		if ((cmd.auxpref & aux_vexpr) != 0) {
			this->STARScmd.Operands[i]->SetBitInSpecFlag4(STARS_VEXPR);
		}

		switch (this->STARScmd.itype) {
		    case NN_vgatherdps:
			case NN_vgatherdpd:
			case NN_vgatherqps:
			case NN_vgatherqpd:
			case NN_vpgatherdd:
			case NN_vpgatherdq:
			case NN_vpgatherqd:
			case NN_vpgatherqq:
			  this->STARScmd.Operands[i]->SetBitInSpecFlag4(STARS_VSIB);
			default:
				;
		}
	}

	return success;
} // end of STARS_IDA_Instruction_t::STARS_GetCmd()

bool STARS_IDA_Instruction_t::OpcodeDefaultsTo64BitOperands(void) const {
	bool Default64;
	switch (this->GetIDAOpcode()) {
		// use ss
	case NN_pop:
	case NN_popf:
	case NN_popfq:
	case NN_push:
	case NN_pushf:
	case NN_pushfq:
	case NN_retn:
	case NN_retf:
	case NN_retnq:
	case NN_retfq:
	case NN_call:
	case NN_callfi:
	case NN_callni:
	case NN_enter:
	case NN_enterq:
	case NN_leave:
	case NN_leaveq:

		// near branches
	case NN_ja:
	case NN_jae:
	case NN_jb:
	case NN_jbe:
	case NN_jc:
	case NN_je:
	case NN_jg:
	case NN_jge:
	case NN_jl:
	case NN_jle:
	case NN_jna:
	case NN_jnae:
	case NN_jnb:
	case NN_jnbe:
	case NN_jnc:
	case NN_jne:
	case NN_jng:
	case NN_jnge:
	case NN_jnl:
	case NN_jnle:
	case NN_jno:
	case NN_jnp:
	case NN_jns:
	case NN_jnz:
	case NN_jo:
	case NN_jp:
	case NN_jpe:
	case NN_jpo:
	case NN_js:
	case NN_jz:
	case NN_jcxz:
	case NN_jecxz:
	case NN_jrcxz:
	case NN_jmp:
	case NN_jmpni:
	case NN_jmpshort:
	case NN_loop:
	case NN_loopq:
	case NN_loope:
	case NN_loopqe:
	case NN_loopne:
	case NN_loopqne:
		Default64 = true;
		break;
	default:
		Default64 = false;
		break;
	}
	return Default64;
} // end of STARS_IDA_Instruction_t::OpcodeDefaultsTo64BitOperands()

bool STARS_IDA_Instruction_t::Has64BitOperands(void) const {
#ifdef __EA64__
	return (((this->STARScmd.auxpref & aux_use64) != 0)
		&& ((this->STARScmd.rex & REX_W) != 0
		|| (((this->STARScmd.auxpref & aux_natop) != 0) && this->OpcodeDefaultsTo64BitOperands())));
	// 64-bit segment, rex.w or insns-64
#else
	return false;
#endif
}

bool STARS_IDA_Instruction_t::Uses32BitAddressing(void) const {
	int Temp = this->STARScmd.auxpref & (aux_use32 | aux_use64 | aux_natad);
	return (Temp == (aux_natad | aux_use32))
		|| (Temp == 0)
		|| (Temp == aux_use64);
}

bool STARS_IDA_Instruction_t::IsUseOpnd(std::size_t OpndNum) const {
   return (this->GetInstFeatures() & UseMacros[OpndNum]);
}

bool STARS_IDA_Instruction_t::IsDefOpnd(std::size_t OpndNum) const {
   return (this->GetInstFeatures() & DefMacros[OpndNum]);
}

// set the USE bit
void STARS_IDA_Instruction_t::SetOpUsed(std::size_t OpndNum) {
	this->STARSfeatures |= UseMacros[OpndNum];
	return;
} 

// reset the USE bit
void STARS_IDA_Instruction_t::SetOpNotUsed(std::size_t OpndNum) {
	this->STARSfeatures &= (~UseMacros[OpndNum]);
	return;
} 

// set the DEF bit
void STARS_IDA_Instruction_t::SetOpDefed(std::size_t OpndNum) {
	this->STARSfeatures |= DefMacros[OpndNum]; 
	return;
}

// reset the DEF bit
void STARS_IDA_Instruction_t::SetOpNotDefed(std::size_t OpndNum) {
	this->STARSfeatures &= (~DefMacros[OpndNum]);
	return;
} 

// Fix up IDA Pro IMUL instruction by removing operand 1
void STARS_IDA_Instruction_t::RemoveIDAOp1ForIMUL(void) {
	this->STARScmd.Operands[1] = this->STARScmd.Operands[2];
	this->STARScmd.Operands[2] = nullptr;
	return;
}

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeVoidOpnd(void) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeVoidOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeImmediateOpnd(STARS_uval_t value) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_imm;
	TempOp.value = value;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeVoidOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeRegOpnd(uint16_t RegNum) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_reg;
	TempOp.dtyp = GetRegDtyp(RegNum, this->Has64BitOperands());
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeFloatingPointRegOpnd(uint16_t RegNum) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_fpreg;
	TempOp.dtyp = GetRegDtyp(RegNum, false);
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeFloatingPointRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMMXRegOpnd(uint16_t RegNum) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_mmxreg;
	TempOp.reg = RegNum;
	TempOp.dtyp = GetRegDtyp(RegNum, false);
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMMXRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeXMMRegOpnd(uint16_t RegNum) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_xmmreg;
	TempOp.reg = RegNum;
	TempOp.dtyp = GetRegDtyp(RegNum, false);
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeXMMRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeYMMRegOpnd(uint16_t RegNum) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_ymmreg;
	TempOp.reg = RegNum;
	TempOp.dtyp = GetRegDtyp(RegNum, false);
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeYMMRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeNearPointerOpnd(STARS_ea_t TargetAddr) const {
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_near;
	TempOp.addr = TargetAddr;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
}

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMemPhraseOpnd(uint16_t BaseRegNum, uint16_t IndexRegNum, uint16_t ScaleFactor) const {
	// TODO: Construct SIB byte when IndexRegNum is used.
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_phrase;
	TempOp.reg = BaseRegNum;
	TempOp.dtyp = GetRegDtyp(BaseRegNum, this->Has64BitOperands());
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMemPhraseOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMemDisplacementOpnd(uint16_t BaseRegNum, uint16_t IndexRegNum, uint16_t ScaleFactor, STARS_ea_t offset) const {
	// TODO: Construct SIB byte when IndexRegNum is used.
	op_t TempOp;
	this->InitOperand(TempOp);
	TempOp.type = o_displ;
	TempOp.reg = BaseRegNum;
	TempOp.addr = (ea_t) offset;
	TempOp.dtyp = GetRegDtyp(BaseRegNum, this->Has64BitOperands());
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMemDisplacementOpnd()

bool STARS_IDA_Instruction_t::IsBranchToFarChunk(SMPInstr *CurrInst, STARS_ea_t &TargetAddr) {
	bool FarBranch = false;
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	func_t *CurrChunk = get_fchunk(this->m_id.GetIDWithinFile());
	for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
		STARSOpndTypePtr JumpTarget = CurrUse->GetOp();
		if (JumpTarget->IsNearPointer() || JumpTarget->IsFarPointer()) {
			// Branches to a code address
			TargetAddr = JumpTarget->GetAddr();
			// stdclib sometimes has jumps to zero and calls to zero. These are dead code.
			if ((0 != TargetAddr) && (STARS_BADADDR != TargetAddr)) {
				func_t *TargetChunk = get_fchunk(TargetAddr);
				// Is target address within the same chunk as the branch?
				FarBranch = (NULL == TargetChunk) || (CurrChunk->startEA != TargetChunk->startEA);
			}
		}
	}
	return FarBranch;
} // end of STARS_IDA_Instruction_t::IsBranchToFarChunk()

